@@ -67,7 +67,7 @@ from . import codec
 from . import const
 from .error import ProtocolError
 from .offset import Offset, OffsetType
-from .types import MessageQueue
+from .types import MessageQueue, PublishError
 from .util import concatv, retry
 
 logger = logging.getLogger(__name__)
@@ -87,6 +87,7 @@ PEER_PROPERTIES = {
 
 DEFAULT_CREDIT = 10
 
+# pylint: disable=no-member
 REQUESTS_KEYS = {
     const.RESPONSE_KEY_PEER_PROPERTIES,    # type: ignore
     const.RESPONSE_KEY_SASL_HANDSHAKE,     # type: ignore
@@ -102,6 +103,7 @@ REQUESTS_KEYS = {
     const.RESPONSE_KEY_CREATE_STREAM,      # type: ignore
     const.RESPONSE_KEY_QUERY_PUBLISHER_SEQUENCE,  # type: ignore
 }
+# pylint: enable=no-member
 
 # TODO: check only for 0x06 code, see https://github.com/rabbitmq/rabbitmq-server/issues/3874
 is_stream_na = lambda ex: ex.code in (2, 6)
@@ -125,6 +127,7 @@ class RabbitMQStreamsProtocol(asyncio.Protocol):
 
     :var transport: Asyncio transport used by the protocol to send and
         receive RabbitMQ streams frames.
+    :var encoder: Frame encoder for RabbitMQ Streams protocol.
     :var decoder: Frame decoder for RabbitMQ Streams protocol.
     """
     transport: Option[asyncio.Transport]
@@ -153,6 +156,17 @@ class RabbitMQStreamsProtocol(asyncio.Protocol):
         self._correlation_id = max(self._requests, default=0)
         self._waiters: dict[int, asyncio.Future[None]] = {}
 
+        #
+        # connection timeout handling
+        #
+
+        # timestamp of last response from rabbitmq streams broker
+        self._timestamp = self._loop.time()
+        # handle of scheduled callback to watch the timestamp
+        self._timestamp_handle: asyncio.TimerHandle | None = None
+        # true when heartbeat sent to check a connection
+        self._heartbeat_sent = False
+
     @property
     def connected(self) -> bool:
         return not self.transport.empty
@@ -483,9 +497,17 @@ class RabbitMQStreamsProtocol(asyncio.Protocol):
     def connection_made(self, transport: asyncio.BaseTransport) -> None:
         self.transport = Option(tp.cast(asyncio.Transport, transport))
 
+        self._heartbeat_sent = False
+        self.schedule_timestamp_callback()
+
     def connection_lost(self, ex: Exception | None) -> None:
         self.transport = Option[asyncio.Transport]()
 
+        if self._timestamp_handle:
+            self._timestamp_handle.cancel()
+            self._timestamp_handle = None
+        self._heartbeat_sent = False
+
         all_tasks = tp.cast(
             tp.Iterable[asyncio.Future[None]],
             concatv(
@@ -511,7 +533,46 @@ class RabbitMQStreamsProtocol(asyncio.Protocol):
         if not self.transport.empty:
             self.transport.value.abort()
 
+    def timestamp_callback(self) -> None:
+        """
+        Callback to watch timing of RabbitMQ Streams broker last response.
+        """
+        heartbeat = self.heartbeat
+        delta = self._loop.time() - self._timestamp
+        if delta > heartbeat and not self._heartbeat_sent:
+            logger.warning(
+                'sending heartbeat, time since last response: {:.1f}'
+                .format(delta)
+            )
+            self.send_frame(codec.HEARTBEAT)
+            self._heartbeat_sent = True
+
+            assert self._heartbeat_sent
+
+        elif delta > heartbeat and self._heartbeat_sent:
+            logger.warning(
+                'stopping connection, time since last response: {:.1f}'
+                .format(delta)
+            )
+            self.abort()
+
+        else:
+            self._heartbeat_sent = False
+
+        self.schedule_timestamp_callback()
+        assert self._heartbeat_sent and delta > heartbeat \
+            or not self._heartbeat_sent and delta <= heartbeat
+
+    def schedule_timestamp_callback(self) -> None:
+        """
+        Callback to watch timing of RabbitMQ Streams broker last response.
+        """
+        self._timestamp_handle = self._loop.call_later(
+            self.heartbeat, self.timestamp_callback
+        )
+
     def data_received(self, chunk: bytes) -> None:
+        self._timestamp = self._loop.time()
         for start, key in self.decoder.commands(chunk):
             # NOTE: data is updated after FrameDecoder.commands call
             data = self.decoder.data
@@ -523,26 +584,13 @@ class RabbitMQStreamsProtocol(asyncio.Protocol):
 
             elif key == const.KEY_HEARTBEAT:
                 # simply wait for a heartbeat from rabbitmq streams broker
-                # and respond with another heartbeat; this seems to work
-                # and active heartbeat sending by client seems to be not
-                # necessary
+                # and respond with another heartbeat; see also
+                # schedule_timestamp_callback method
                 self.send_frame(codec.HEARTBEAT)
                 logger.info('heartbeat frame sent')
 
             elif key == const.KEY_TUNE:
-                # decode tune command and accept what rabbitmq streams
-                # broker provides
-                frame_size, heartbeat = codec.FMT_TUNE.unpack_from(data, start)[-2:]
-                self.frame_size = frame_size
-                self.heartbeat = heartbeat
-                self.encode = codec.Encoder(frame_size)
-
-                # tune request is sent by rabbitmq streams broker; respond
-                # to the request with another tune request (not a response)
-                self.send_frame(data[start:start + codec.FMT_TUNE.size])
-                task = self._waiters.pop(key)
-                task.set_result(None)
-                logger.info('tune frame sent')
+                self.process_tuning(data, start)
 
             elif key == const.RESPONSE_KEY_CREDIT:  # type: ignore
                 code, subscription_id = codec.decode_credit(data, start)
@@ -563,12 +611,8 @@ class RabbitMQStreamsProtocol(asyncio.Protocol):
                 self.process_request_response(key, data, start)
 
             elif key == const.KEY_PUBLISH_ERROR:
-                publisher_id, errors = codec.decode_publish_error(data, start)
-                logger.warning(
-                    'publish error; publisher={}, publishing id=0x{:x},' \
-                    ' error=0x{:04x}, num errors={}'.format(
-                        publisher_id, *errors[0], len(errors)
-                ))
+                error = codec.decode_publish_error(data, start)
+                self.process_publish_error(error)
             else:
                 logger.warning('unknown key; key=0x{:04x}'.format(key))
 
@@ -680,6 +724,25 @@ class RabbitMQStreamsProtocol(asyncio.Protocol):
         else:
             assert False, 'wrong assumptions about message ids'
 
+    def process_publish_error(self, error: PublishError) -> None:
+        """
+        Process publishing errors.
+        """
+        publisher_id, errors = error
+        match errors:
+            case ((publishing_id, err), *_):
+                logger.warning(
+                    'publish error; publisher={}, publishing id=0x{:x},' \
+                    ' error=0x{:04x}, num errors={}'.format(
+                        publisher_id, publishing_id, err, len(errors)
+                ))
+            case _:
+                logger.warning(
+                    'empty publish error; publisher={}'.format(
+                        publisher_id,
+                    )
+                )
+
     def process_message_delivery(self, data: bytes, start: int) -> None:
         """
         Process message delivery from RabbitMQ Streams broker.
@@ -705,4 +768,33 @@ class RabbitMQStreamsProtocol(asyncio.Protocol):
                 'subscription not found, id={}'.format(sid)
             )
 
+    def process_tuning(self, data: bytes, start: int) -> None:
+        """
+        Process RabbitMQ Streams broker tunning response.
+
+        Decode tune command received from RabbitMQ Streams broker and
+        accept tunning parameters it provides.
+
+        :param data: Data received from RabbitMQ Streams broker.
+        :param start: Data starting point.
+        """
+        assert const.KEY_TUNE in self._waiters
+
+        frame_size, heartbeat = codec.FMT_TUNE.unpack_from(data, start)[-2:]
+        self.frame_size = frame_size
+        self.heartbeat = heartbeat
+        self.encoder = codec.Encoder(frame_size)
+
+        # tune request is sent by rabbitmq streams broker; respond
+        # to the request with another tune request (not a response)
+        self.send_frame(data[start:start + codec.FMT_TUNE.size])
+        task = self._waiters.pop(const.KEY_TUNE)
+        task.set_result(None)
+        logger.info(
+            'tune frame sent, frame size={}, heartbeat={}'
+            .format(self.frame_size, self.heartbeat)
+        )
+
+        assert const.KEY_TUNE not in self._waiters
+
 # vim: sw=4:et:ai
